#include "string.h"

namespace optimized_string
{

String::String()
{
	this->m_pString = new PascalString();
	this->m_Type = PASCAL_STRING;
}

String::String(const String& Other)
:	m_pString(NULL)
{
	this->Set(Other);
}

String::String(const char* szString)
:	m_pString(NULL)
{
	this->Set(szString);
}

String::~String()
{
	if (this->m_pString)
	{
		delete this->m_pString;
	}
}

void String::Reset() 
{
	if (this->m_pString)
	{
		delete this->m_pString;
		this->m_pString = NULL;
	}
}

void String::Set(const String& Other)
{
	this->Reset();

	// CONSTRUCTION SET
	if (Other.m_Type == PASCAL_STRING)
	{
		const PascalString* pOther;
		pOther = reinterpret_cast<const PascalString*>(Other.m_pString);
		this->m_pString = new PascalString(*pOther);
		this->m_Type = PASCAL_STRING;
	}
	else
	{
		const CString* pOther;
		pOther = reinterpret_cast<const CString*>(Other.m_pString);
		this->m_pString = new CString(*pOther);
		this->m_Type = C_STRING;
	}
}

void String::Set(const char* szString)
{
	this->Reset();

	size_t Size = strlen(szString);
	if (Size > PascalString::MAX_LENGTH)
	{
		// MAKE C-String
		this->m_pString = new CString(szString, Size);
		this->m_Type = C_STRING;
	}
	else
	{
		this->m_pString = new PascalString(szString, 
										   static_cast<unsigned short>(Size));
		this->m_Type = PASCAL_STRING;
	}
}

void String::Append(const String& Other)
{
	size_t Len = Other.Len();

	if (this->m_Type == PASCAL_STRING)
	{
		PascalString* pThis = reinterpret_cast<PascalString*>(this->m_pString);
		if ((Len + pThis->Len()) <= PascalString::MAX_LENGTH)
		{
			if (Other.m_Type == PASCAL_STRING)	
			{
				const PascalString* pOther;
				pOther = reinterpret_cast<const PascalString*>(Other.m_pString);
				pThis->Append(*pOther, Len);
			}
			else
			{
				const CString* pOther;
				pOther = reinterpret_cast<const CString*>(Other.m_pString);
				pThis->Append(pOther->GetCString(), Len);
			}
		}
		else
		{
			// CONVERSION
			const char* szOtherString = NULL;
			if (Other.m_Type == PASCAL_STRING)	
			{
				const PascalString* pOther;
				pOther = reinterpret_cast<const PascalString*>(Other.m_pString);
				szOtherString = pOther->GetString();
			}
			else
			{
				const CString* pOther;
				pOther = reinterpret_cast<const CString*>(Other.m_pString);
				szOtherString = pOther->GetCString();
			}

			size_t NewLen = pThis->Len() + Len;
			char* szNewString = new char[NewLen + 1];
			memcpy(szNewString, 
				   pThis->GetString(), 
				   sizeof(char) * pThis->Len());
			memcpy(&szNewString[pThis->Len()], 
				   szOtherString, 
				   sizeof(char) * Len);
			szNewString[NewLen] = 0;

			this->Reset();
			this->m_pString = new CString(szNewString, NewLen);
			this->m_Type = C_STRING;
		}
	}
	else
	{
		CString* pThis = reinterpret_cast<CString*>(this->m_pString);
		if (Other.m_Type == PASCAL_STRING)	
		{
			const PascalString* pOther;
			pOther = reinterpret_cast<const PascalString*>(Other.m_pString);
			pThis->Append(pOther->GetString(), Len);
		}
		else
		{
			const CString* pOther;
			pOther = reinterpret_cast<const CString*>(Other.m_pString);
			pThis->Append(*pOther, Len);
		}
	}
}

void String::Append(const char* szString)
{
	size_t Len = strlen(szString);

	if (this->m_Type == PASCAL_STRING)
	{
		PascalString* pThis = reinterpret_cast<PascalString*>(this->m_pString);
		if ((Len + pThis->Len()) <= PascalString::MAX_LENGTH)
		{
			pThis->Append(szString, Len);
		}
		else
		{
			// CONVERSION
			size_t NewLen = pThis->Len() + Len;
			char* szNewString = new char[NewLen + 1];
			memcpy(szNewString, 
				   pThis->GetString(), 
				   sizeof(char) * pThis->Len());
			memcpy(&szNewString[pThis->Len()], szString, sizeof(char) * Len);
			szNewString[NewLen] = 0;

			this->Reset();
			this->m_pString = new CString(szNewString, NewLen);
			this->m_Type = C_STRING;
		}
	}
	else
	{
		CString* pThis = reinterpret_cast<CString*>(this->m_pString);
		pThis->Append(szString, Len);
	}
}

const bool String::Compare(const String& Other) const
{
	if (!this->m_pString)
	{
		return false;
	}
	else if (this->m_Type == PASCAL_STRING)
	{
		if (Other.m_Type != PASCAL_STRING)
		{
			return false;
		}
		else
		{
			const PascalString* pThis;
			pThis = reinterpret_cast<const PascalString*>(this->m_pString);
			const PascalString* pOther;
			pOther = reinterpret_cast<const PascalString*>(Other.m_pString);
			return pThis->Compare(*pOther);
		}
	}
	else if (this->m_Type == C_STRING)
	{
		const CString* pThis;
		pThis = reinterpret_cast<const CString*>(this->m_pString);
		const CString* pOther;
		pOther = reinterpret_cast<const CString*>(Other.m_pString);
		return pThis->Compare(*pOther);
	}

	return false;
}

const bool String::Compare(const char* szString) const
{
	if (!this->m_pString)
	{
		return false;
	}
	else if (this->m_Type == PASCAL_STRING)
	{
		size_t Len = strlen(szString);
		if (Len < PascalString::MAX_LENGTH)
		{
			const PascalString* pThis;
			pThis = reinterpret_cast<const PascalString*>(this->m_pString);
			return pThis->Compare(szString, Len);
		}
		else
		{
			return false;
		}
	}
	else if (this->m_Type == C_STRING)
	{
		size_t Len = strlen(szString);
		if (Len < PascalString::MAX_LENGTH)
		{
			return false;
		}
		else
		{
			const CString* pThis;
			pThis = reinterpret_cast<const CString*>(this->m_pString);
			return pThis->Compare(szString, Len);
		}
	}

	return false;
}

void String::LTrim()
{
	if (!this->m_pString)
		return;

	if (this->m_Type == PASCAL_STRING)
	{
		PascalString* pThis = reinterpret_cast<PascalString*>(this->m_pString);
		pThis->LTrim();
	}
	else if (this->m_Type == C_STRING)
	{
		CString* pThis = reinterpret_cast<CString*>(this->m_pString);
		pThis->LTrim();
	}
}

void String::RTrim()
{
	if (!this->m_pString)
		return;

	if (this->m_Type == PASCAL_STRING)
	{
		PascalString* pThis = reinterpret_cast<PascalString*>(this->m_pString);
		pThis->RTrim();
	}
	else if (this->m_Type == C_STRING)
	{
		CString* pThis = reinterpret_cast<CString*>(this->m_pString);
		pThis->RTrim();
	}
}

size_t String::Len()
{
	if (!this->m_pString)
		return 0;

	if (this->m_Type == PASCAL_STRING)
	{
		PascalString* pThis = reinterpret_cast<PascalString*>(this->m_pString);
		return pThis->Len();
	}
	else if (this->m_Type == C_STRING)
	{
		CString* pThis = reinterpret_cast<CString*>(this->m_pString);
		return pThis->Len();
	}

	return 0;
}

const size_t String::Len() const
{
	if (!this->m_pString)
		return 0;

	if (this->m_Type == PASCAL_STRING)
	{
		const PascalString* pThis;
		pThis = reinterpret_cast<const PascalString*>(this->m_pString);
		return pThis->Len();
	}
	else if (this->m_Type == C_STRING)
	{
		const CString* pThis;
		pThis = reinterpret_cast<const CString*>(this->m_pString);
		return pThis->Len();
	}

	return 0;
}

const char* String::GetCString() 
{
	if (!this->m_pString)
		return NULL;

	if (this->m_Type == PASCAL_STRING)
	{
		PascalString* pThis = reinterpret_cast<PascalString*>(this->m_pString);
		return pThis->GetCString();
	}
	else if (this->m_Type == C_STRING)
	{
		CString* pThis = reinterpret_cast<CString*>(this->m_pString);
		return pThis->GetCString();
	}

	return NULL;
}

};